#include <NewSoftSerial.h>

#define CMD_SIZE 6
// Maximum allowed errors when reading picture data
#define MAX_ERRORS 15
// The byte offset from the end of a data package where
// JPEG image data ends
#define PACKAGE_DATA_END_OFFSET 2
// How long to wait for camera to process JPEG data
#define PROCESS_DELAY 1000
// Number of sync attempts to try before giving up
#define MAX_SYNC_ATTEMPTS 60
// The byte offset where image data starts in a JPEG image
// data package
#define PACKAGE_DATA_START 4
// How long to wait for serial communication responses
#define RESPONSE_DELAY 100

// command bytes
static const byte CMD_PREFIX = 0xAA;
static const byte CMD_ACK = 0x0E;
static const byte CMD_DATA = 0x0A;
static const byte CMD_GETPICTURE = 0x04;
static const byte CMD_INITIAL = 0x01;
static const byte CMD_LIGHTFREQ = 0x13;
static const byte CMD_PACKAGESIZE = 0x06;
static const byte CMD_SNAPSHOT = 0x05;
static const byte CMD_SYNC = 0x0D;

// ColorTypes
static const byte CT_GRAYSCALE_2 = 0x01; 
static const byte CT_GRAYSCALE_4 = 0x02; 
static const byte CT_GRAYSCALE_8 = 0x03; 
static const byte CT_COLOR_12 = 0x05;
static const byte CT_COLOR_16 = 0x06; 
static const byte CT_JPEG = 0x07;

// Frequency Types
static const byte FT_50Hz = 0x00;
static const byte FT_60Hz = 0x01;

//JPEGResolution
static const byte JR_80x64 = 0x01;
static const byte JR_160x128 = 0x03;
static const byte JR_320x240 = 0x05;
static const byte JR_640x480 = 0x07;

//Picture Type
static const byte PT_SNAPSHOT = 0x01;
static const byte PT_PREVIEW = 0x02;
static const byte PT_JPEG = 0x05;

//PreviewResolution
static const byte PR_80x60 = 0x01;
static const byte PR_160x120 = 0x03;

//Snapshot Type
static const byte ST_COMPRESSED = 0x00;
static const byte ST_UNCOMPRESSED = 0x01;

static const byte LAST_JPEG_ACK = 0xF0F0;



byte _command[CMD_SIZE];
byte _receive_cmd[CMD_SIZE];

NewSoftSerial camera(12,6);

void setup() {
  delay(3000);
  Serial.begin(9600);
  camera.begin(9600);
  Serial.println("Starting sync...");
  delay(10000);
  if(!sync()) {
    Serial.println("Sync failed");
  } 
  if(!initial(CT_JPEG, PR_160x120, JR_160x128)) {
    Serial.println("Init failed");  
  } 
  if(!setPackageSize(64)) {
    Serial.println("Package size failed"); 
  } 
  if(!setLightFrequencyType(FT_50Hz)) {
    Serial.println("Light frequency failed"); 
  } 
  delay(2000);
  if(!snapshot(ST_COMPRESSED,0)) {
    Serial.println("Snapshot failed");
  }
  delay(100); 
  if(!getPicture(PT_JPEG)) {
    Serial.println("Get picture failed");
  } 
}

void loop() {

}

void createCommand(byte cmd, byte param1, byte param2, byte param3, byte param4 ) {
  _command[0] = CMD_PREFIX;
  _command[1] = cmd;
  _command[2] = param1;
  _command[3] = param2;
  _command[4] = param3;
  _command[5] = param4;
}

boolean getPicture(byte pictureType) {
  int errorCount = 0;
  int pictureSize = 0;
  // Keeps track of the number of packages received
  int packageCount = 0;
  // Current byte position in the ENTIRE image
  int pictureDataPosition = 0;
  // The size of the picture ID within the package
  int packagePictureID = 0;
  // The size of the picture data within the package
  int packagePictureDataSize = 0;
  createCommand(CMD_GETPICTURE, pictureType, 0x00, 0x00, 0x00);
  sendCommand();
  receiveReply();
  if(_receive_cmd[1] == CMD_ACK && _receive_cmd[2] == CMD_GETPICTURE) {
    receiveReply();
    if(_receive_cmd[1] == CMD_DATA) {
      pictureSize = _receive_cmd[5] << 8;
      pictureSize |= _receive_cmd[4] << 8;
      pictureSize |= _receive_cmd[3];
      if (pictureSize < 0) {
        Serial.println("Error receiving pictureSize...");
        return false; 
      }
      Serial.print("Picture size: ");
      Serial.print(pictureSize, DEC);
      Serial.print('\n'); 
    } 
    else {
      return false; 
    }
    //I'm using an '@' character to signal to the receiver that picture data is on the way...it's my own protocol
    Serial.print('@');
    // Everything sent to the Serial line after this should be JPEG picture data only unless there is an error
    // picture processing loop
    while( pictureDataPosition < pictureSize && errorCount < MAX_ERRORS) {
      createCommand(CMD_ACK, 0x00, 0x00, (byte)(packageCount & 0xFF), (byte)(packageCount >> 8));
      sendCommand();
      byte package_header[PACKAGE_DATA_START];
      if(!waitForResponse(PROCESS_DELAY, package_header, sizeof(package_header))) {
        return false;
      }
      // This is the package ID
      packagePictureID = package_header[1] << 8;
      packagePictureID |= package_header[0];
      // Size of the data payload
      packagePictureDataSize = package_header[3] << 8;
      packagePictureDataSize |= package_header[2];
      // Read the image data
      byte package_data[packagePictureDataSize];
      if(!waitForResponse(PROCESS_DELAY, package_data, sizeof(package_data))) {
        return false;
      }
      // Read the checksum
      byte chksum_data[PACKAGE_DATA_END_OFFSET];
      if(!waitForResponse(PROCESS_DELAY, chksum_data, sizeof(chksum_data))) {
        return false;
      }
      // calculate the checksum
      byte chksum_calc = 0;
      for (int i = 0; i < sizeof(package_header) ; i++)
      {
        chksum_calc += package_header[i];
      }
      for (int i = 0; i < sizeof(package_data); i++)
      {
        chksum_calc += package_data[i];
      }
      if (chksum_calc == chksum_data[0])
      {
        pictureDataPosition += packagePictureDataSize;
        for(int i = 0; i < sizeof(package_data); i++) {
          Serial.print(package_data[i],BYTE); 
        }
        packageCount++;
      } 
      else {
        //invalid checksum try again
        errorCount++;
      }
    } // end picture processing loop
    // Send last ACK
    createCommand(CMD_ACK, LAST_JPEG_ACK, 0x00, (byte)(packageCount & 0xFF), (byte)(packageCount >> 8));
    sendCommand();
    if (errorCount < MAX_ERRORS) {
      return true;
    } 
    else {
      return false; 
    }
  } 
  else {
    return false;
  }
  return true;
}

boolean initial(byte colorType, byte previewResolution, byte jpegResolution) {
  createCommand(CMD_INITIAL, 0x00, colorType, previewResolution, jpegResolution);
  sendCommand();
  if( !receiveReply() )
  {
    Serial.println("Error receiving reply from initial...");
    return false; 
  } 
  if(_receive_cmd[1] == CMD_ACK && _receive_cmd[2] == CMD_INITIAL) {
    return true;
  } 
  else {
    return false; 
  }
}

boolean receiveReply() {
  delay(100);
  if( camera.available() > 0 )
  {
    for(int i = 0;i < 6; i++) {
      _receive_cmd[i] = camera.read(); 
    }
    return true;
  } 
  else {
    return false; 
  }
}



void sendCommand() {
  for(int i = 0; i < 6; i++) {
    camera.print(_command[i],BYTE); 
  }
}

boolean setLightFrequencyType(byte frequencyType) {
  createCommand(CMD_LIGHTFREQ, frequencyType, 0x00, 0x00, 0x00);
  sendCommand();
  if( !receiveReply() )
  {
    Serial.println("Error receiving reply from setLightFrequency...");
    return false; 
  } 
  if(_receive_cmd[1] == CMD_ACK && _receive_cmd[2] == CMD_LIGHTFREQ) {
    return true;
  } 
  else {
    return false;
  }
}

boolean setPackageSize(int size) {
  createCommand(CMD_PACKAGESIZE, 0x08, (byte)(size & 0xFF), (byte)(size >> 8), 0x00);
  sendCommand();
  if( !receiveReply() )
  {
    Serial.println("Error receiving reply from setPackageSize...");
    return false; 
  } 
  if(_receive_cmd[1] == CMD_ACK && _receive_cmd[2] == CMD_PACKAGESIZE) {
    return true;
  } 
  else {
    return false;
  }
}

boolean snapshot(byte snapshotType, int skipFrames) {
  createCommand(CMD_SNAPSHOT, snapshotType, (byte)(skipFrames & 0xFF), (byte)(skipFrames >> 8), 0x00);
  sendCommand();  
  if( !receiveReply() )
  {
    Serial.println("Error receiving reply from snapshot...");
    return false;
  }
  if(_receive_cmd[1] == CMD_ACK && _receive_cmd[2] == CMD_SNAPSHOT) {
    return true;
  } 
  else {
    return false;
  }
}

boolean sync() {
  bool success = false;
  int attempts = 0;
  // manual indicates we might need 60 attempts at syncing
  while(attempts < 60) {
    createCommand(CMD_SYNC, 0x00, 0x00, 0x00, 0x00);
    sendCommand();
    receiveReply();
    if(_receive_cmd[1] == CMD_ACK && _receive_cmd[2] == CMD_SYNC) {
      receiveReply();
      if(_receive_cmd[1] == CMD_SYNC) {
        camera.flush();
        createCommand(CMD_ACK, CMD_SYNC, 0x00, 0x00, 0x00);
        sendCommand();
        return true;
      } 
      else {
        return false; 
      }
    }
    attempts++;
  }
  return success;
}



boolean waitForResponse( int timeout, byte buffer[], int bufferLength )
{
  int byteCnt = 0;
  unsigned long time = millis();

  while( millis() - time <= timeout )
  {
    while( camera.available() > 0 )
    {
      buffer[byteCnt] = camera.read();
      byteCnt++;
      if( byteCnt == bufferLength )
      {
        return true;
      }
    }
  }
  if( byteCnt > 0 )
  {
    return true;
  }
  return false;
}
















































